﻿namespace U2UConsult.EF4.Linq.Tips
{
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Some Entity Framework 4.0 Linq tips.
    /// </summary>
    public class Program
    {
        /// <summary>
        /// Where it all starts.
        /// </summary>
        /// <param name="args">Command line arguments.</param>
        public static void Main(string[] args)
        {
            // Tip 1: Only select the needed columns.
            // SelectAllColumns();
            // ProjectColumns();

            // Tip 2: Avoid the 'Include' syntax
            // JoinWithInclude();
            // JoinInLinq();
            // JoinThroughProjection();

            // Tip 3: Consider alternatives
            // U2UConsult.L2S.Linq.Program.L2SOuterJoin();
            // JoinInSql();

            // Tip 4: Stick to EF40, but try harder
            // JoinInLinqWithProjection();
            // U2UConsult.L2S.Linq.Program.L2SOuterJoinWithProjection();

            Console.ReadKey();
        }

        /// <summary>
        /// Fetches all People as fully populated entities.
        /// </summary>
        private static void SelectAllColumns()
        {
            List<Person> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People
                         select p).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}, {1}", person.LastName, person.FirstName));
            }
        }

        /// <summary>
        /// Fetches all People as Data Transfer Objects carrying only the needed properties.
        /// </summary>
        private static void ProjectColumns()
        {
            List<PersonDto> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People
                         select new PersonDto() { LastName = p.LastName, FirstName = p.FirstName }).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}, {1}", person.LastName, person.FirstName));
            }
        }

        /// <summary>
        /// Fetches all People with their Phones, using the EF4 'Include'.
        /// </summary>
        private static void JoinWithInclude()
        {
            List<Person> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People.Include("PersonPhones")
                         select p).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}: {1}", person.LastName, person.PersonPhones.First().PhoneNumber));
            }
        }

        /// <summary>
        /// Fetches all People with their Phones, using a native Linq left outer join.
        /// </summary>
        private static void JoinInLinq()
        {
            List<PersonWithPhonesDto> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People
                         join pp in model.PersonPhones
                         on p.BusinessEntityID equals pp.BusinessEntityID
                         into phones
                         select new PersonWithPhonesDto() { LastName = p.LastName, PersonPhones = phones }).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}: {1}", person.LastName, person.PersonPhones.First().PhoneNumber));
            }
        }

        /// <summary>
        /// Fetches all People with their Phones, using a native Linq projection.
        /// </summary>
        private static void JoinThroughProjection()
        {
            List<PersonWithPhonesDto> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People
                         select new PersonWithPhonesDto() { LastName = p.LastName, PersonPhones = p.PersonPhones }).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}: {1}", person.LastName, person.PersonPhones.First().PhoneNumber));
            }
        }

        /// <summary>
        /// Fetches all People with their Phones, using a native TSQL query.
        /// </summary>
        private static void JoinInSql()
        {
            string joinStatement = @"SELECT [t0].[LastName], [t1].[PhoneNumber]
		                               FROM [Person].[Person] AS [t0]
                            LEFT OUTER JOIN [Person].[PersonPhone] AS [t1] 
			                             ON [t1].[BusinessEntityID] = [t0].[BusinessEntityID]";

            List<PersonWithPhoneDto> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = model.ExecuteStoreQuery<PersonWithPhoneDto>(joinStatement).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}: {1}", person.LastName, person.PhoneNumber));
            }
        }

        //
        // And the winner is:
        //

        /// <summary>
        /// Fetches all People with their Phones.
        /// </summary>
        private static void JoinInLinqWithProjection()
        {
            List<PersonWithPhoneDto> query;

            using (AdventureWorks2008Entities model = new AdventureWorks2008Entities())
            {
                query = (from p in model.People
                         from pp in
                             (from zz in model.PersonPhones
                              where p.BusinessEntityID == zz.BusinessEntityID
                              select new { zz.PhoneNumber }).DefaultIfEmpty()
                         select new PersonWithPhoneDto() { LastName = p.LastName, PhoneNumber = pp.PhoneNumber }).ToList();
            }

            foreach (var person in query)
            {
                Console.WriteLine(string.Format("{0}: {1}", person.LastName, person.PhoneNumber));
            }
        }

        //
        // DTO zone
        //

        /// <summary>
        /// Person as a Data Transfer Object.
        /// </summary>
        private class PersonDto
        {
            public string LastName { get; set; }

            public string FirstName { get; set; }
        }

        /// <summary>
        /// Person-with-phones as a Data Transfer Object.
        /// </summary>
        private class PersonWithPhonesDto
        {
            public string LastName { get; set; }

            public IEnumerable<PersonPhone> PersonPhones { get; set; }
        }

        /// <summary>
        /// Person-with-phone as a Data Transfer Object.
        /// </summary>
        private class PersonWithPhoneDto
        {
            public string LastName { get; set; }

            public string PhoneNumber { get; set; }
        }
    }
}
